import { doesExist } from '@apextoaster/js-utils';
import { stdin as input, stdout as output } from 'node:process';
import { createInterface, Interface as Readline } from 'node:readline/promises';

import { CommandContext } from '../command/index.js';
import { catchAndLog } from '../utils/async.js';
import { matchCommands } from '../utils/command.js';
import { LOCAL_DISCRIMINATOR } from '../utils/roles.js';
import { Author, Bot, BotContext, Channel, Message } from './index.js';

export const READLINE_AUTHOR: Author = {
  discriminator: LOCAL_DISCRIMINATOR,
  username: 'readline',
};

export async function onLine(context: CommandContext, channel: Channel, rl: Readline, line: string): Promise<void> {
  const { args, commands, logger } = context;
  logger.debug({ line }, 'message received');

  const matchingCommands = matchCommands(line, commands, args.commandPrefix);
  for (const command of matchingCommands) {
    logger.debug({
      command: command.name,
    }, 'message matched command');

    if (doesExist(command.auth)) {
      const allowed = await command.auth(context, READLINE_AUTHOR);
      logger.debug({ allowed, author: READLINE_AUTHOR, command }, 'command auth check');

      if (allowed === false) {
        logger.warn({ author: READLINE_AUTHOR, command }, 'author attempted to use unauthorized command');
        continue;
      }
    }

    const message: Message = {
      author: READLINE_AUTHOR,
      channel,
      content: line,
      async reply(reply: string) {
        logger.info({ reply }, 'reply from command');
        return this;
      },
    };

    try {
      await command.run(context, message);
    } catch (err) {
      if (err instanceof Error) {
        logger.error(err, 'error executing command');
      } else {
        logger.error({ err }, 'unknown error executing command');
      }
    } finally {
      rl.prompt();
    }
  }
}

/**
 * Start the readline bot interface.
 *
 * This reads input from stdin and send replies through the logging system, and is meant for interactive
 * debugging.
 *
 * @public
 */
export async function readlineConnect(botContext: BotContext, readlineFactory: typeof createInterface = createInterface): Promise<Bot> {
  const { logger } = botContext;
  const rl = readlineFactory({ input, output });

  const channel: Channel = {
    name: 'readline',
    type: 'dm',
  };

  const bot: Bot = {
    async connect() {
      rl.prompt();
      return true;
    },
    destroy() {
      rl.close();
    },
    async getChannel() {
      return channel;
    },
    async getUser() {
      return READLINE_AUTHOR;
    },
  };

  rl.on('close', () => {
    logger.info('readline bot closed');
    // process.kill(process.pid, 'SIGINT');
  });

  rl.on('line', (line) => catchAndLog(onLine({
    ...botContext,
    bot,
  }, channel, rl, line), logger, 'error handling line'));

  rl.setPrompt('> ');

  return bot;
}
