import { signal, SIGNAL_RESET } from '@apextoaster/js-utils';
import { createLogger, DEBUG } from 'bunyan';

import { APP_NAME, parseArgs } from './args.js';
import { Bot, BotContext, BOTS } from './bot/index.js';
import { COMMANDS } from './command/index.js';
import { DATA_LAYERS } from './data/index.js';
import { startCrons } from './job/cron.js';
import { rconConnect } from './utils/rcon.js';

/**
 * Process exit codes.
 */
export enum ExitCode {
  SUCCESS = 0,
  ERROR = 1,
}

/**
 * Connect to some bots and wait for commands or an exit signal.
 */
export async function main(argv: Array<string>): Promise<ExitCode> {
  const args = await parseArgs(argv);

  if (args.help === true) {
    return ExitCode.SUCCESS;
  }

  const logger = createLogger({
    name: APP_NAME,
    level: DEBUG,
  });

  logger.info({app: APP_NAME, args}, 'starting bot');

  for (const command of COMMANDS) {
    logger.debug({ command }, 'command debug');
  }

  logger.info({ data: args.data }, 'connecting to data store');
  const dataCtor = DATA_LAYERS[args.data];
  const data = new dataCtor(args);
  await data.connect();

  logger.info({ host: args.rconHost, port: args.rconPort }, 'connecting to rcon server');
  const rcon = await rconConnect(args, logger);
  await rcon.connect();

  logger.info('starting bots');
  const bots: Array<Bot> = [];
  const botContext: BotContext = {
    args,
    commands: COMMANDS,
    data,
    logger,
    rcon,
  };

  for (const botName of args.bots) {
    const ctor = BOTS[botName];
    if (typeof ctor === 'function') {
      logger.info({ botName }, 'starting bot');
      const bot = await ctor(botContext);
      await bot.connect();
      bots.push(bot);
    } else {
      logger.warn({ botName }, 'unknown bot');
    }
  }

  logger.debug('starting cron jobs');
  const cron = await startCrons({
    args,
    bot: bots[0],
    data,
    logger,
    rcon,
  });

  await signal(SIGNAL_RESET);

  logger.info('stopping cron jobs');
  await cron.destroy();

  logger.info('stopping bots');
  for (const bot of bots) {
    bot.destroy();
  }

  rcon.end();

  return ExitCode.SUCCESS;
}
