import yargs, { PositionalOptions } from 'yargs';

import { Author } from '../bot/index.js';
import { Command } from '../command/index.js';

/**
 * Friendly name for an Author.
 *
 * @public
 */
export function authorName(author: Author): string {
  return `${author.username}#${author.discriminator}`;
}

/**
 * Wrapper for append with a default prefix.
 *
 * @public
 */
export function commandName(command: Command, prefix: string): string {
  return `${prefix}${command.name}`;
}

/**
 * Match command names against a message body.
 *
 * @public
 */
export function matchCommands(message: string, commands: ReadonlyArray<Command>, prefix: string): ReadonlyArray<Command> {
  const [head, ..._rest] = message.split(' ');

  return commands.filter((command) => {
    const name = commandName(command, prefix);
    return name === head;
  });
}

/**
 * Parse and remove the command name from a message body.
 *
 * @public
 */
export function removeCommandName(message: string, command: Command, prefix: string): string {
  const name = commandName(command, prefix);
  const [head, ...rest] = message.split(' ');

  if (head === name) {
    return rest.join(' ');
  } else {
    return message;
  }
}

/**
 * Filter out string types.
 *
 * @public
 */
export type StringsOnly<T> = T extends string ? T : never;

/**
 * Transform a type into its yargs options.
 *
 * @public
 */
export type BodyOptions<T> = { [key in StringsOnly<keyof T>]: yargs.Options };

/**
 * Remove the command name from a message, parse any CLI args, and return them with the remainder of the message body.
 *
 * @public
 */
// eslint-disable-next-line max-params
export async function parseBody<
  ArgT,
  OptionT extends BodyOptions<ArgT> = BodyOptions<ArgT>,
  ArgVT = ArgT & { _: Array<string> }
>(
  message: string,
  command: Command,
  prefix: string,
  options: OptionT,
  positionals: Array<[StringsOnly<keyof ArgT>, PositionalOptions]> = []
): Promise<[string, ArgVT]> {
  const body = removeCommandName(message, command, prefix);
  // eslint-disable-next-line @typescript-eslint/ban-types
  let parser: yargs.Argv<yargs.Omit<{}, keyof OptionT>> = yargs()
    .exitProcess(false)
    .options<OptionT>(options)
    .help();

  for (const [key, positionalOptions] of positionals) {
    parser = parser.positional(key, positionalOptions);
  }

  const args = await parser.parse(body);

  return [
    args._.join(' '),
    args as unknown as ArgVT, // TODO: hax, the compiler says there is no overlap
  ];
}
