/* eslint-disable @typescript-eslint/strict-boolean-expressions, no-invalid-this */
import { BaseError, Logger } from 'noicejs';
import { Rcon } from 'rcon-client';
import rconPacket, { Packet } from 'rcon-client/lib/packet.js';

import { ParsedArgs } from '../args.js';

/**
 * RCON client methods.
 *
 * @public
 */
export interface RconClient {
  connect(): Promise<this>;
  end(): void;
  send(command: string): Promise<string>;
}

/**
 * Secret interface used by the monkey-patch methods that allows access to protected fields of the RCON client that
 * are marked as private in the types.
 *
 * @internal
 */
interface RconClientPrivates {
  connect(): Promise<this>;
  handlePacket(data: Buffer): void;

  authenticated: boolean;
  callbacks: Map<number, (packet: Packet) => void>;
  requestId: number;
}

/**
 * Create (and monkey-patch) an RCON client.
 *
 * @public
 */
export async function rconConnect(args: ParsedArgs, logger: Logger, rconCtor: typeof Rcon = Rcon): Promise<RconClient> {
  const client = new rconCtor({
    host: args.rconHost,
    port: args.rconPort,
    password: args.rconPassword,
    timeout: 5000,
  });

  /**
   * Monkey patch for https://github.com/janispritzkau/rcon-client/issues/21
   *
   * The server packet numbering repeats 0 or something, causing a mismatch and timeout.
   */
  function handlePacket(this: RconClientPrivates, data: Buffer) {
    const packet = rconPacket.decodePacket(data);
    const id = this.authenticated ? (packet.id + 1) : (this.requestId - 1);

    logger.debug({
      id,
      packetId: packet.id,
      packetType: packet.type,
    }, 'handle rcon packet');

    const handler = this.callbacks.get(id);
    if (handler) {
      handler(packet);
      this.callbacks.delete(id);
    }
  };

  Reflect.set(client, 'handlePacket', handlePacket);

  const originalConnect = Reflect.get(client, 'connect');

  /**
   * Super hax: the server resets the packet ID when reconnecting, causing an authentication error.
   */
  function connect(this: RconClientPrivates) {
    this.requestId = 0;

    return originalConnect.call(this);
  }

  Reflect.set(client, 'connect', connect);

  client.on('error', (err: unknown) => {
    if (err instanceof Error) {
      logger.error(err, 'rcon client error');
    } else {
      logger.error({ err }, 'unknown rcon client error');
    }

    client.connect().then(() => {
      logger.info('rcon reconnected');
    }, (connectError) => {
      logger.error(connectError, 'rcon unable to reconnect');
    });
  });

  return client;
}

export const RETRY_MESSAGE = 'Not connected';

/**
 * Send an RCON command and attempt to reconnect and retry if it fails.
 *
 * @public
 */
export async function sendWithRetry(client: RconClient, command: string, logger: Logger, retries = 1): Promise<string> {
  try {
    return await client.send(command);
  } catch (err) {
    if (err instanceof Error) {
      logger.error(err, 'error sending rcon command');
      if (err.message === RETRY_MESSAGE && retries > 0) {
        logger.debug({ command, retries }, 'retrying rcon command');
        await client.connect();
        return sendWithRetry(client, command, logger, retries - 1);
      } else {
        throw new BaseError('error sending command', err);
      }
    } else {
      logger.error({ err }, 'unknown error sending rcon command');
      throw err;
    }
  }
}
