import { expect } from 'chai';
import { ChannelManager, ChannelType, Client, ClientOptions, ClientUser, Message, User } from 'discord.js';
import { NullLogger } from 'noicejs';
import { createStubInstance, stub } from 'sinon';

import { parseArgs } from '../../src/args.js';
import { discordConnect, onMessage } from '../../src/bot/discord.js';
import { Command, CommandContext } from '../../src/command/index.js';
import { ping } from '../../src/command/ping.js';
import { MemoryDataLayer } from '../../src/data/memory.js';
import { mockBot, mockRcon, REQUIRED_ARGS } from '../helpers.js';

describe('discord bot', () => {
  it('should watch for the client ready event', async () => {
    const args = await parseArgs(REQUIRED_ARGS);
    const rcon = mockRcon();

    const destroyMock = stub<[], void>();
    class MockDiscord<Ready extends boolean> extends Client<Ready> {
      constructor(options: ClientOptions) {
        super(options);

        return createStubInstance<Client<Ready>>(Client, {
          destroy: destroyMock,
          on: stub(),
          once: stub(),
        });
      }
    }

    const bot = await discordConnect({
      args,
      commands: [],
      data: new MemoryDataLayer(args),
      logger: NullLogger.global,
      rcon,
    }, MockDiscord);

    bot.destroy();
    expect(destroyMock).to.have.callCount(1);
  });

  it('should watch for messages and match commands', async () => {
    const args = await parseArgs(REQUIRED_ARGS);
    const bot = mockBot({});
    const rcon = mockRcon();

    const discord: Client = createStubInstance<Client<true>>(Client, {
      destroy: stub(),
      on: stub(),
      once: stub(),
    });
    discord.channels = createStubInstance(ChannelManager);
    discord.channels.fetch = stub().resolves({
      guild: {
        name: '',
      },
      id: 'test',
      type: ChannelType.GuildText,
    });
    const user: ClientUser = createStubInstance(ClientUser);
    discord.user = user;

    const context: CommandContext = {
      args,
      bot,
      commands: [ping],
      data: new MemoryDataLayer(args),
      logger: NullLogger.global,
      rcon,
    };

    const author: User = createStubInstance(User);
    const message: Message = createStubInstance(Message);
    message.author = author;
    message.content = '!ping';

    await onMessage(context, discord, message);

    // eslint-disable-next-line @typescript-eslint/unbound-method
    expect(message.reply).to.have.callCount(1);
  });

  it('should authenticate commands', async () => {
    const command: Command = {
      name: 'test',
      desc: '',
      auth: stub().resolves(false),
      run: stub().resolves(),
    };

    const args = await parseArgs(REQUIRED_ARGS);
    const bot = mockBot({});
    const rcon = mockRcon();

    const discord: Client = createStubInstance<Client<true>>(Client, {
      destroy: stub(),
      on: stub(),
      once: stub(),
    });
    discord.channels = createStubInstance(ChannelManager);
    discord.channels.fetch = stub().resolves({
      guild: {
        name: '',
      },
      id: 'test',
      type: ChannelType.GuildText,
    });
    const user: ClientUser = createStubInstance(ClientUser);
    discord.user = user;

    const context: CommandContext = {
      args,
      bot,
      commands: [command],
      data: new MemoryDataLayer(args),
      logger: NullLogger.global,
      rcon,
    };

    const author: User = createStubInstance(User);
    const message: Message = createStubInstance(Message);
    message.author = author;
    message.content = '!test';

    await onMessage(context, discord, message);

    // eslint-disable-next-line @typescript-eslint/unbound-method
    expect(message.reply).to.have.callCount(0);
  });

  it('should ignore its own messages');
  it('should get author info based on user snowflake');
});
