import { expect } from 'chai';
import { NullLogger } from 'noicejs';
import { Rcon } from 'rcon-client/lib/index.js';
import { mock, stub } from 'sinon';

import { parseArgs } from '../../src/args.js';
import { RconClient, rconConnect, RETRY_MESSAGE, sendWithRetry } from '../../src/utils/rcon.js';
import { REQUIRED_ARGS } from '../helpers.js';

describe('rcon utils', () => {
  describe('rcon client', () => {
    it('should attempt to reconnect on errors', async () => {
      const args = await parseArgs(REQUIRED_ARGS);
      const client = await rconConnect(args, NullLogger.global);

      const connectStub = stub(client, 'connect').resolves(client);

      (client as unknown as Rcon).emitter.emit('error', new Error('test'));

      expect(connectStub).to.have.callCount(1);
    });
  });

  describe('send with retry helper', () => {
    it('should send the command for each retry', async () => {
      const retryCount = 2;
      const sendCount = retryCount + 1;

      const sendMock = mock().atLeast(sendCount).atMost(sendCount).named('RconClient.send').rejects(new Error(RETRY_MESSAGE));
      const client: RconClient = {
        connect: mock().atLeast(1).named('RconClient.connect').resolves(),
        end: mock().named('RconClient.end').resolves(),
        send: sendMock,
      };

      await expect(sendWithRetry(client, 'test', NullLogger.global, retryCount)).to.eventually.be.rejectedWith();

      expect(sendMock).to.have.callCount(sendCount);
    });

    it('should call connect before each retry', async () => {
      const retryCount = 1;
      const sendCount = retryCount + 1;

      const connectMock = mock().named('RconClient.connect').resolves();
      const sendMock = mock().atLeast(sendCount).atMost(sendCount).named('RconClient.send').rejects(new Error(RETRY_MESSAGE));
      const client: RconClient = {
        connect: connectMock,
        end: mock().named('RconClient.end').resolves(),
        send: sendMock,
      };

      await expect(sendWithRetry(client, 'test', NullLogger.global, retryCount)).to.eventually.be.rejectedWith();

      expect(connectMock).to.have.been.calledBefore(sendMock);
    });
  });
});
